/*
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 * Copyright 1993 by David Dawes <dawes@xfree86.org>
 * Copyright 2002 by SuSE Linux AG, Author: Egbert Eich
 * Copyright 1994-2002 by The XFree86 Project, Inc.
 * Copyright 2002 by Paul Elliott
 * (Ported from xf86-input-mouse, above copyrights taken from there)
 * Copyright © 2008 University of South Australia
 * Copyright © 2010 by Elmar Hanlhofer http://www.plop.at
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of the authors
 * not be used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The authors make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/* Touch screen code. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <X11/Xatom.h>
#include <xf86.h>
#include <xf86Xinput.h>
#include <exevents.h>

#include <evdev-properties.h>
#include "evdev.h"

enum {
    RBEMU_DISABLED = 0,
    RBEMU_ENABLED,
    RBEMU_AUTO
};

#ifdef HAVE_PROPERTIES
static Atom prop_rbemu     = 0; /* Right button emulation on/off property */
static Atom prop_rbtimeout = 0; /* Right button timeout property */
#endif


int
EvdevTouchTimer(InputInfoPtr pInfo)
{
    EvdevPtr pEvdev = pInfo->private;
    int	sigstate;

    sigstate = xf86BlockSIGIO ();

    pEvdev->touchscreen.rightbuttonpending = FALSE;
    pEvdev->touchscreen.buttonstateR = 1;

    pEvdev->touchscreen.touchfilterpending = FALSE;
    pEvdev->touchscreen.buttonstateL = 0;

    EvdevPostButtonEvent(pInfo, 3, 1);

    if (pEvdev->debug)
        xf86Msg(X_INFO, "%s: Right button press\n",
                pInfo->name);


    xf86UnblockSIGIO (sigstate);
    return 0;
}

int
EvdevTouchFlushLeftButton(InputInfoPtr pInfo)
{
    EvdevPtr pEvdev = pInfo->private;
    int	sigstate;

    if (!pEvdev->touchscreen.touchfilterpending)
	return 0;

    sigstate = xf86BlockSIGIO ();

    pEvdev->touchscreen.touchfilterpending = FALSE;
    pEvdev->touchscreen.rightbuttonpending = FALSE;	
    pEvdev->touchscreen.buttonstateR = 0;

    if (pEvdev->touchscreen.buttonstateL == 0)
    {

	EvdevPostButtonEvent(pInfo, 1, 1);
    
	if (pEvdev->debug)
    	    xf86Msg(X_INFO, "%s: Filter, left button press\n", pInfo->name);
    }
    
    EvdevPostButtonEvent(pInfo, 1, 0);
    
    if (pEvdev->debug)
        xf86Msg(X_INFO, "%s: Filter, left button release\n", pInfo->name);

    pEvdev->touchscreen.buttonstateL = 0;

    xf86UnblockSIGIO (sigstate);
    return 0;
}


/**
 * Button press.
 *
 * @param code button number (1 for left, 3 for right)
 * @param press TRUE if press, FALSE if release.
 *
 * @return TRUE if event was swallowed, FALSE
 * otherwise.
 */
BOOL
EvdevTouchFilterEvent(InputInfoPtr pInfo, int button, BOOL press)
{
    EvdevPtr pEvdev = pInfo->private;
    int ret = FALSE;


    if (!pEvdev->touchscreen.enabled)
        return ret;


    /* don't care about other buttons */
    if (button != 1)
        return ret;
	
    if (pEvdev->debug)
        xf86Msg(X_INFO, "%s: Check buttons\n", pInfo->name);

    if ((!press) && (pEvdev->touchscreen.buttonstateR == 1) && (pEvdev->touchscreen.rightbuttonenabled))
    {
	EvdevQueueButtonEvent(pInfo, 3, 0);

	if (pEvdev->debug)
    	    xf86Msg(X_INFO, "%s: Release right button\n", pInfo->name);
	
	pEvdev->touchscreen.rightbuttonpending = FALSE;
	pEvdev->touchscreen.buttonstateR = 0;
	pEvdev->touchscreen.buttonstateL = 0;
	pEvdev->touchscreen.touchfilterpending = FALSE;
    } 
    else if ((!press) && (pEvdev->touchscreen.buttonstateR == 0) && (pEvdev->touchscreen.touchfilterenabled))
    {
	if (pEvdev->debug)
    	    xf86Msg(X_INFO, "%s: Start touch filter timeout\n", pInfo->name);
	
        pEvdev->touchscreen.touchfilterexpires = GetTimeInMillis () + pEvdev->touchscreen.touchfiltertimeout;
        pEvdev->touchscreen.touchfilterpending = TRUE;
    }
    else if ((pEvdev->touchscreen.touchfilterpending) && (press) && (pEvdev->touchscreen.touchfilterenabled))
    {
        pEvdev->touchscreen.touchfilterpending = FALSE;

	if (pEvdev->debug)
    	    xf86Msg(X_INFO, "%s: Touch filter hit, do not send button press/release\n", pInfo->name);
    }
    else if ((!pEvdev->touchscreen.rightbuttonpending) && (press) && (pEvdev->touchscreen.rightbuttonenabled))
    {
	if (pEvdev->debug)
	    xf86Msg(X_INFO, "%s: Start right button timeout\n", pInfo->name);

        pEvdev->touchscreen.rightbuttonexpires = GetTimeInMillis () + pEvdev->touchscreen.rightbuttontimeout;
        pEvdev->touchscreen.rightbuttonpending = TRUE;
        pEvdev->touchscreen.pressed_x = pEvdev->vals[0];
        pEvdev->touchscreen.pressed_y = pEvdev->vals[1];
    }
    else {
	
	if (pEvdev->touchscreen.buttonstateL == 0)
	{
	    pEvdev->touchscreen.buttonstateL = 1;

	    EvdevQueueButtonEvent(pInfo, 1, 1);

	    if (pEvdev->debug)
	        xf86Msg(X_INFO, "%s: Left button press\n", pInfo->name);
	}

	if (!press)
	{
	    pEvdev->touchscreen.buttonstateL = 0;
	    EvdevQueueButtonEvent(pInfo, 1, 0);
		
	    if (pEvdev->debug)
	        xf86Msg(X_INFO, "%s: Left button release\n", pInfo->name);
	}

    	pEvdev->touchscreen.rightbuttonpending = FALSE;	
	pEvdev->touchscreen.buttonstateR = 0;
    }
	
    return TRUE;
}


void EvdevTouchWakeupHandler(pointer data,
                             int i,
                             pointer LastSelectMask)
{
    InputInfoPtr pInfo = (InputInfoPtr)data;
    EvdevPtr     pEvdev = (EvdevPtr)pInfo->private;
    int ms;

    if (pEvdev->touchscreen.rightbuttonpending)
    {
        ms = pEvdev->touchscreen.rightbuttonexpires - GetTimeInMillis();
        if (ms <= 0)
            EvdevTouchTimer(pInfo);
    }

    if (pEvdev->touchscreen.touchfilterpending)
    {
        ms = pEvdev->touchscreen.touchfilterexpires - GetTimeInMillis();
        if (ms <= 0)
            EvdevTouchFlushLeftButton(pInfo);
    }

}

void EvdevTouchBlockHandler(pointer data,
                            struct timeval **waitTime,
                            pointer LastSelectMask)
{
    InputInfoPtr    pInfo = (InputInfoPtr) data;
    EvdevPtr        pEvdev= (EvdevPtr) pInfo->private;
    int             ms;

    if (pEvdev->touchscreen.rightbuttonpending)
    {
        ms = pEvdev->touchscreen.rightbuttonexpires - GetTimeInMillis ();
        if (ms <= 0)
            ms = 0;
        AdjustWaitForDelay (waitTime, ms);
    }

    if (pEvdev->touchscreen.touchfilterpending)
    {
        ms = pEvdev->touchscreen.touchfilterexpires - GetTimeInMillis ();
        if (ms <= 0)
            ms = 0;
        AdjustWaitForDelay (waitTime, ms);
    }
}

void
EvdevTouchPreInit(InputInfoPtr pInfo)
{
    EvdevPtr pEvdev = (EvdevPtr)pInfo->private;

    pEvdev->touchscreen.enabled = TRUE;
    
    pEvdev->touchscreen.rightbuttonpending = FALSE;
    pEvdev->touchscreen.touchfilterpending = FALSE;

    pEvdev->touchscreen.rightbuttonenabled = xf86SetBoolOption(pInfo->options, "RightClick", TRUE);
    xf86Msg(X_INFO, "%s: RightClick mouse button emulation %s.\n",
            pInfo->name, (pEvdev->touchscreen.rightbuttonenabled) ? "on" : "off");

    pEvdev->touchscreen.rightbuttontimeout = xf86SetIntOption(pInfo->options, "RightClickMS", 2000);
    pEvdev->touchscreen.rightclicksquare = xf86SetIntOption(pInfo->options, 
					    "RightClickSquare", 0) / 2;

    pEvdev->touchscreen.touchfilterenabled = xf86SetBoolOption(pInfo->options, "TouchFilter", TRUE);
    pEvdev->touchscreen.touchfiltertimeout = xf86SetIntOption(pInfo->options, "TouchFilterMS", 10);

}

void
EvdevTouchOn(InputInfoPtr pInfo)
{
    if (!pInfo->dev->button) /* don't init for keyboards */
        return;

    RegisterBlockAndWakeupHandlers (EvdevTouchBlockHandler,
                                    EvdevTouchWakeupHandler,
                                    (pointer)pInfo);
}

void
EvdevTouchFinalize(InputInfoPtr pInfo)
{
    if (!pInfo->dev->button) /* don't cleanup for keyboards */
        return;

    RemoveBlockAndWakeupHandlers (EvdevTouchBlockHandler,
                                  EvdevTouchWakeupHandler,
                                  (pointer)pInfo);

}

/* Enable/disable right mouse button emulation. */
void
EvdevTouchEnable(InputInfoPtr pInfo, BOOL enable)
{
    EvdevPtr pEvdev = (EvdevPtr)pInfo->private;
    if (pEvdev->touchscreen.rightbuttonenabled == RBEMU_AUTO)
        pEvdev->touchscreen.rightbuttonenabled = enable;
}


#ifdef HAVE_PROPERTIES
static int
EvdevTouchSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
                      BOOL checkonly)
{
    InputInfoPtr pInfo  = dev->public.devicePrivate;
    EvdevPtr     pEvdev = pInfo->private;

    if (atom == prop_rbemu)
    {
        if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
            return BadMatch;

        if (!checkonly)
            pEvdev->touchscreen.rightbuttonenabled = *((BOOL*)val->data);
    } else if (atom == prop_rbtimeout)
    {
        if (val->format != 32 || val->size != 1 || val->type != XA_INTEGER)
            return BadMatch;

        if (!checkonly)
            pEvdev->touchscreen.rightbuttontimeout = *((CARD32*)val->data);
    }

    return Success;
}

/**
 * Initialise property for RB emulation on/off.
 */
void
EvdevTouchInitProperty(DeviceIntPtr dev)
{
    InputInfoPtr pInfo  = dev->public.devicePrivate;
    EvdevPtr     pEvdev = pInfo->private;
    int          rc;

    if (!dev->button) /* don't init prop for keyboards */
        return;

    prop_rbemu = MakeAtom(EVDEV_PROP_MIDBUTTON, strlen(EVDEV_PROP_MIDBUTTON), TRUE);
    rc = XIChangeDeviceProperty(dev, prop_rbemu, XA_INTEGER, 8,
                                PropModeReplace, 1,
                                &pEvdev->touchscreen.rightbuttonenabled,
                                FALSE);
    if (rc != Success)
        return;
    XISetDevicePropertyDeletable(dev, prop_rbemu, FALSE);

    prop_rbtimeout = MakeAtom(EVDEV_PROP_MIDBUTTON_TIMEOUT,
                              strlen(EVDEV_PROP_MIDBUTTON_TIMEOUT),
                              TRUE);
    rc = XIChangeDeviceProperty(dev, prop_rbtimeout, XA_INTEGER, 32, PropModeReplace, 1,
                                &pEvdev->touchscreen.rightbuttontimeout, FALSE);

    if (rc != Success)
        return;
    XISetDevicePropertyDeletable(dev, prop_rbtimeout, FALSE);

    XIRegisterPropertyHandler(dev, EvdevTouchSetProperty, NULL, NULL);
}
#endif
